%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS assert.ql,v 1.6 1993/12/10 02:23:10 marku Exp


assert_thy(Name, List) :-
	assert_thy_body(Name, List, true).

assert_thy(CD, Name, List) :-
	assert_thy_body(CD, Name, List, true).

assert_thy_body(Name, List, Body) :- 
	once((
	        loading(CD)
	    ;
	        current_dir(CD)
	    )),
	assert_thy_body(CD, Name, List, Body), !.

assert_thy_body(Name, List, Body) :-
	assert_global_body(Name, List, Body), !.

assert_thy_body(Name, List, Body) :-
	writelist(['Assert failed for ', Name, ' : ', List, nl]), fail.

assert_thy_body(CD, Name, List, Body) :-
	length(List, N),
	(saved_fact(Name/N) -> update_dir(CD) ; local_fact(Name/N)),
	\+  \+  (thaw(List),
		 thaw(Body),
		 assertz(([Name, CD] : List :- Body))), !.

assert_thy_body(CD, Name, List, Body) :-
	assert_global_body(Name, List, Body).

assert_global(Name, List) :-
	assert_global_body(Name, List, true).

assert_global_body(Name, List, Body) :-
	length(List, N),
	global_fact(Name/N),
	update_dir(global), 
        \+  \+  (thaw(List),
		 thaw(Body),
		 assertz((Name : List :- Body))), !.

assert_global_body(Name, List, Body) :-
	writelist(['Assertglobal failed for ', Name, ' : ', List, nl]), fail.

retract_thy(CD, Name, List) :-
        length(List, N),
        retract([Name, CD] : List), !,
        update_dir(CD).


update_dir(CD) :- updated_dir(CD), !.
update_dir(CD) :- assert(updated_dir(CD)).

/*
** Tactic versions: restrict to 'extra' fact types
*/


tactic_assert(Name, List) :-
	once((
	        loading(CD)
	    ;
	        current_dir(CD)
	    )),
        tactic_assert(CD, Name, List), !.

tactic_assert(Name, List) :-
        tactic_assert_global(Name, List).

tactic_assert(CD, Name, List) :-
        length(List, N),
	((theory_saved_fact : [Name/N]) -> 
		update_dir(CD)
	    ; 
		(theory_local_fact : [Name/N])),
	!,
        \+  \+  (thaw(List), assertz([Name, CD] : List)).

tactic_assert(CD, Name, List) :-
        tactic_assert_global(Name, List).

tactic_assert_global(Name, List) :-
        length(List, N),
        (extra_global_fact : [Name/N]),
        update_dir(global),
        \+  \+  (thaw(List), assertz(Name : List)), !.

tactic_retract(CD, Name, List) :-
        length(List, N),
	theory_extra_fact(Name/N),
        retract([Name, CD] : List), !,   
        update_dir(CD).

tactic_retract(Name, List) :-
	length(List, N),
	global_extra_fact(Name/N),
	retract(Name:List), !,
	update_dir(global).
	


/*
** Force a save
*/

touch(CD) :-
 	\+  (loaded(CD),
	     is_theory_name(CD),
	     \+  updated_dir(CD),
             \+  assert(updated_dir(CD))).

