%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS win_rule.ql,v 1.5 1993/12/10 02:23:22 marku Exp


?- op(810,xfy,(and)).
?- op(830,xfy,(=>)).

?- object_var(x).

openrule(Term, Pos, NewHyps, ShiftHyp, ShiftEqu, ShiftGoal) :-
    fix_path(Term, Pos, CPos, SubTerm),
    var(SubTerm),
    check_term(Term),
    check_terms(NewHyps),
    check_ruleequ(ShiftEqu),
    expand_id(ShiftHyp, OtNH),
    expand_id(ShiftGoal, OtNG),
    replace_subterm(Term, CPos, F, LF),
    assert_thy(openrule, [Term, CPos, SubTerm, NewHyps, OtNH, ShiftEqu, OtNG, (F --> LF)]).


expand_id(id, [(A --> A)]).
expand_id([], []).
expand_id([id|B], [(A --> A)|B]).
expand_id([(A --> B)|C], [(A --> B)|D]) :-
	expand_id(C, D).


add_antecedant([], C, C).
add_antecedant([A|B], C, AB => C) :-
	list_to_and([A|B], AB).


list_to_and([], true).
list_to_and([X], X) :- !.
list_to_and([H|T], H and AT) :- list_to_and(T, AT).


/*
** Neat little routine
*/

place(N, L, El) :-
	length([_|Pre], N),
	append(Pre, [El|_], L).

    
/*
** Get the subterm, obvars, and fixed path from the focus + path
*/



fix_path(Term, [], [], Term).

fix_path(Term, [P|Pos], [b(X)|NewPos], Subterm) :-
        quantify(Term, Quant, X, Body), !,
        once(( (P = 2) ; (P = b(X)) )),
        fix_path(Body, Pos, NewPos, Subterm).

fix_path(Term, [P|Pos], [P|NewPos], Subterm) :-
        Term =.. [Head|Arguments], !,
        place(P, Arguments, Arg),
        fix_path(Arg, Pos, NewPos, Subterm).


/*
** From xtrans (of all places)  (with mod's)
*/

replace_subterm(T, [], A, A).

replace_subterm(T, [Pos|Tail], A, NewT) :-
    T =.. TList, !,
    length(PreList, Pos),
    append(PreList, [Arg|PostList], TList),
    replace_subterm(Arg, Tail, A, NewArg),
    append(PreList, [NewArg|PostList], NewTList),
    NewT =.. NewTList.

replace_subterm(T, [b(_)|Tail], A, NewT) :-
    quantify(   T, Q, V,    B), !,
    replace_subterm(B, Tail, A, NewB),
    quantify(NewT, Q, V, NewB).

