%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS prove.ql,v 1.13 1993/12/10 02:23:47 marku Exp


?- op(810,xfy,(and)).
?- op(830,xfy,(=>)).


/*
** Prove that a term is implied by true
*/

lemma(Name, Term)         :- prove(Name, Term, lemma,   fail).
theorem(Name, Term)       :- prove(Name, Term, theorem, true).
lemma(Name, Term, Tact)   :- prove(Name, Term, lemma,   fail, Tact).
theorem(Name, Term, Tact) :- prove(Name, Term, theorem, fail, Tact).



/*
** Construct an appropriate tactic call
*/

prove(Name, F, Type, Ask) :-
    (interactive_tactic : [Tact]), 
    do_prove(Name, F, Type, Ask, Tact).

prove(Name, F, Type, Ask, Tact) :-
    (interactive_tactic : [Call]), 
    Code =.. [Call, Tact],    
    do_prove(Name, F, Type, Ask, Code).



/* 
** Universal proof banana
*/

do_prove(DName, F, Type, Ask, Tact) :-
    \+ \+ (
	once((
	      decode_rule(DName, CD, Name),
	      check_atom(Name),
	      cd(CD),
	      available(Name),
              add_proof_name(CD : Name),

              ipSet(newvars,      []),
              ipSet(proofvars,    []),
              ipSet(newvar_flag,  true),

	      check_term(F),
	      promote_var(F),
	      
	      equal_to(rev(=>), E),

	      ipSet(abandon_proof_record,  false),
	      ipSet(safe_hyps,	  false),
              ipSet(ask_theorem,  Ask),
              ipSet(proof_type,   Type),
	      ipSet(verbosity,    1),

	      ipSet(cw,        1),
	      ipSet(wrap,      1),
	      ipSet(in_hyp,    false),
	      ipSet(focuslist, [(F, initial_focus)]),
	      ipSet(focusdep,  []),
	      ipSet(goal,      true),
	      ipSet(relation,  E),
	      ipSet(hyplist,   []),
	      ipSet(newhyps,   0),
	      ipSet(hypnr,     0)
	      )),

	once(Tact),

    	once((
	    complete_window(E, Proof, true, _),

	    get_confirmation(Type, (CD : Name), F, Proof),
	    save_proof_record(CD:Name, F, Proof),
	    assert_thy(CD, Type, [Name, F])
	    ))
	).


get_confirmation(_, _, _, _) :-
	ipLookup(ask_theorem, fail), !.

get_confirmation(Type, Name, Term, Proof) :-
	writelist(['The resultant ', Type, ' (', Name, '):', nl,
		'    ', Term, nl, nl]),
	question('Do you wish to keep this result? ').

/*
** Check that we aren't proving the same theorem twice
*/

add_proof_name(N) :-
	ipLookup(proof_names, X),
        check_or_writelist(
		\+  member(N, X), 
		['There is a proof in progress with this name.']),
	ipSet(proof_names, [N|X]).



/*
** Save the proof record to a file.
*/


save_proof_record(Theory:Name, Term, Proof) :-
	ipLookup(abandon_proof_record, false), !,
	proof_file_name(Theory:Name, File),
	once((
	        save_to(File)
	    ;
	        writeln('\n Creating proofs directory ...'),
		system(mkdir, [mkdir, Theory], _), 
		save_to(File)
	    )),
	writeq(proof_of(Name, Term, '1.3', Proof)), writeln(' :-'),
	ipLookup(newvars, NV),
	ipLookup(proofvars, PV),
	get_not_free_goals((NV, PV), NFI),
	write_nfi_list(NFI),
	saved.

save_proof_record(Name, Term, Proof).


write_nfi_list([]) :-
	writeln('true.').
write_nfi_list([A|B]) :-	
	writeq(A), writeln(','),
	write_nfi_list(B).

/*
** Query the proof state
*/

is_goal(Goal)   :- ipLookup(goal,      Goal).
is_reln(R)      :- ipLookup(relation,  R).
is_focus(Focus) :- ipLookup(focuslist, [(Focus, _)|_]).

ask_to_store_theorem      :- ipLookup(ask_theorem, on).
ask_to_store_theorem(on)  :- ipSet(ask_theorem, on).
ask_to_store_theorem(off) :- ipSet(ask_theorem, off).

in_proof :- ipLookup(proof_type, T), T \= fail.


initialize_proof_structure :-
        ipSet(cw, 0),
        ipSet(focuslist, []),
        ipSet(goal, fail),
        ipSet(relation,  fail),
        ipSet(hyplist, []),
        ipSet(hypnr, 0),
        ipSet(proof_names, []),
        ipSet(proof_type, fail).

