%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS prinz.ql,v 1.8 1993/12/10 02:18:58 marku Exp


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Some tactics for the 'r' theory.
%%  By Andreas Prinz, UQ, Sept 1993.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%% The following tactic adds a new hypothesis.
%% You have to prove the correctness of this hypothesis first

:- tactic(new_hyp(_,_),
	 ['new_hyp(+Hyp) -- prove ''Hyp'', then add it as a hypothesis'],
         ['Proving a new hypothesis'],
         ['']
         ).

new_hyp(Hyp,[HProof]):-
   take_theorem(r:refleximp,(Hyp => Hyp), Num),
   open_hyp(Num, 
	(open_pos([1], (delete_new_hyps(Num), HProof)), 
	 trans(r:trueimp))),                            %% KW - simplified
   !.



%% The next tactic introduces an abbreviation.
%% Actually, it only introduces the equality X=abbrev,
%% which you can use as you please.

:- tactic(abbreviate(Term,_),
	'abbreviate(+Term) -- introduce a variable to abbreviate ''Term''',
	['Abbreviating ',Term],
	['']
        ).

abbreviate(Abbv, [Goon]):-
   is_focus(F), trans(r:ifparallel,r,if((X=Abbv),F,F),[]),
   promote_var(X), !, open_pos([2], Goon), !, demote_var(X),
   open_pos([1], trans_true(r:reflexeql)), trans(r:iftrue).



%% A tactic to do case analysis
%% A list of cases is given, or only one case, that
%% is completed with its opposite.
%% Cases can also be entered by an or-clause, but note
%% in this case, that or is left-assoc!
%% Only the second case is searched for more or's.
%%
%% First, the completeness of the cases has to be proven,
%% then the cases are done one by one.
%%
%% If an of the form 'h X' is given, the appropriate hypotheses
%% is taken as definition of cases. Similarily, if the argument
%% is a single atom, a theorem or axiom with this name is
%% taken as cases.

:- tactic(cases(X,_),
	['cases(+X) -- proof by case analysis.',nl,
	 '  If X is a single predicate, the cases are X and not X.',nl,
	 '  X can also be a list of predicates, a disjunction, or',nl,
	 '  the name of a hypothesis or theorem that is a disjunction'],
	['Proof by cases: ', X],
        [['Proof of completeness'],['Case ', 1] | _]
        ).


cases([H|T],C):- 
   !, transform_to_or([H|T],Cases), cases(Cases, C).
cases(h X,[_|C]):- 
   !, is_hyp(X,Hyp), !, cases(Hyp, [trans_true(h X)|C]).
cases(Thm,[_|C]):-
   atom(Thm), !, take_theorem(Thm,T,Num), delete_hyp(Num), !,
   cases(T, [trans_true(Thm)|C]).
cases(One,[_|C]):- 
   \+  (One = (_ or _)), !, cases(One or not One, [trans_true(ornot)|C]).
cases((Case1 or Case2), [C1|R]):-
   is_focus(F), number_of_hyps(Num),
   trans(r:back_thm4,f,
      if((Case1 or Case2),((Case1 => F) and (Case2 => F)),F),[r:caseor]), !,
   open_pos([1], (delete_new_hyps(Num), call_tac(C1))), !, trans(iftrue),
   cases(R), call_tac(_, 'End of Case Analysis', true).

cases([First,Next|R]):-
   nonvar(First), !, is_focus((_ => _) and (_ => _)),
   First = call_tac(- N, Descr, _),
   Next = call_tac(- N, NDescr, _),
   var(NDescr),
   next_label(Descr, NDescr), !,
   do_cases([First, Next | R]).
cases(L):- do_cases(L).

do_cases([Code | Rest]):-
   is_focus((_ => _) and (Case2 => _)),
   number_of_hyps(Num),
   open_pos([1], 
     (delete_new_hyps(Num), !,
      open_pos([2], call_tac(Code)), !,
      trans_true(r:imp_true)) ),
   trans(r:trueand), !,
   two_cases(Case2,Rest).

two_cases(C1 or C2, Code):-
   is_focus((C1 or C2) => F), !,
   open_pos([2], 
     (is_hyp(Num, C1 or C2),
      trans(r:caseor,f,((C1 => F) and (C2 => F)),[Num]), 
      delete_hyp(Num),
      cases(Code))), !,
   trans_true(r:imp_true).
two_cases(Last, [Code|_]) :-
   is_focus(Last => F), !,
   open_pos([2], call_tac(Code)), !,
   trans_true(r:imp_true).

transform_to_or([A],A).
transform_to_or([H|T],H or TT):- transform_to_or(T,TT).



:- tactic(use_ex(Thm,_),
	 ['use_ex(ThmOrHyp) -- use an existential fact.',nl,
	  'This takes a copy of ThmOrHyp, which must be of the',nl,
	  'form (ex x A), and strips off the ex quantifier.',nl,
	  'If ThmOrHyp is uninstantiated, it uses any hypothesis'],
	 ['Stripping ex from ',Thm],
	 ['']
	 ).

use_ex(Name,[X]):-
    ((var(Name) ->
        is_hyp(FullName, (ex x _)))
    ;
	th_or_hyp(Name, EX, FullName)
    ),
    EX = ex x E,
    trans_dir(r:mp,r,[FullName]),
    trans_dir(r:all_dist_imp_r,r,[]),
    !,
    focus([2,2],X).

